import os
from tokre.setup import setup
from tokre.core.parsing import parse, escape
from tokre.core.tree_to_module import tree_to_module, compile
from tokre.core.synth_feat import SynthFeat, collect_matches

from tokre.core.modules import EmbedData, PredData, PartialMatch, Mixer
from tokre.core.pyregex import pyregex_literal

from tokre.labelling.create_label import create_label, get_all_toks
from tokre.labelling.literal_set_utils import load_literal_set, save_literal_set
from tokre.labelling.get_words import get_word_counts

# See https://stackoverflow.com/a/35904211/10222613
# I think this avoids import cycle issues
import sys
this = sys.modules[__name__]

this._tokenizer = None
this._workspace = None
this._openai_api_key = os.environ.get("OPENAI_API_KEY", None)


def get_tokenizer():
    if this._tokenizer is None:
        raise ValueError(
            "Tokenizer not initialized. Call `tokre.setup(tokenizer=your_tokenizer)` first."
        )
    return this._tokenizer


def get_workspace(assert_exists=True):
    if this._workspace is None and assert_exists is True:
        raise ValueError(
            "Workspace not initialized. Call `tokre.setup(tokenizer=your_tokenizer, workspace='your_workspace')` first."
        )
    return this._workspace

import transformers
import tokenizers

def encode(text):
    tokenizer = get_tokenizer()
    
    enc_kwargs = {'add_special_tokens': False}\
        if isinstance(tokenizer, (transformers.PreTrainedTokenizer, transformers.PreTrainedTokenizerFast, tokenizers.Tokenizer))\
        else {}

    return tokenizer.encode(text, **enc_kwargs)


enc = encode


def decode(tokens):
    tokenizer = get_tokenizer()
    return tokenizer.decode(tokens)


dec = decode


def tok_split(s: str):
    tok_ids = encode(s)
    tok_strs = [decode([tok_id]) for tok_id in tok_ids]
    return tok_strs




from tokre.core.modules import Embed, Mixer, VarRef, PartialMatch, randstr, toks_eq
from torch import nn


def tok_split(s):
    tok_ids = tokre.enc(s)
    return [tokre.dec([tok_id]) for tok_id in tok_ids]


def get_literal_variants(tok_literal: list[str]):
    literal_str = ("".join(tok_literal)).strip()
    variants = [tok_split(literal_str), tok_split(" " + literal_str)]
    return variants


class BEGIN(nn.Module):
    def __init__(self):
        super().__init__()
        self.name = f"BEGIN:{randstr()}"

    def matches(self, toks, partial, reversed):
        if toks[partial.end] == "[BEGIN]":
            return [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end + 1,
                    defns=partial.defns,
                    data=None,
                )
            ]
        else:
            return []

class AbsPos(nn.Module):
    def __init__(self, max_pos_idx=129):
        super().__init__()
        self.name = f"pos:{randstr()}"
        self.pos_embed = Embed(max_pos_idx)

    def matches(self, toks, partial, reversed):
        return [
            PartialMatch(
                name=self.name,
                start=partial.end,
                end=partial.end,
                defns=partial.defns,
                data=self.pos_embed(partial.end),
            )
        ]



class VarVariant(nn.Module):
    def __init__(self, var_ref):
        super().__init__()
        assert isinstance(var_ref, VarRef), var_ref
        self.name = f"VarVariant:{var_ref}:{randstr()}"
        self.var_name = var_ref.var_name

    def matches(self, toks, partial, reversed):
        res = []
        if self.var_name in partial.defns:
            var_toks = partial.defns[self.var_name]
            variants = get_literal_variants(var_toks)
            for variant in variants:
                if toks_eq(toks[partial.end : partial.end + len(variant)], variant):
                    res.append(
                        PartialMatch(
                            name=self.name,
                            start=partial.end,
                            end=partial.end + len(variant),
                            defns=partial.defns,
                            data=None,
                        )
                    )
        return res



class VarVariantPrefix(nn.Module):
    def __init__(self, var_ref, max_len=128):
        super().__init__()
        assert isinstance(var_ref, VarRef) or isinstance(var_ref, str), var_ref
        self.name = f"VarVariantPrefix:{str(var_ref)}:{randstr()}"
        
        self.var_name = str(var_ref) if isinstance(var_ref, str) else var_ref.var_name
        self.var_len_and_prefix_idx = Embed((max_len + 1, max_len + 1))
        self.max_len = max_len

    def matches(self, toks, partial, reversed):
        res = []
        if self.var_name in partial.defns:
            var_toks = partial.defns[self.var_name]
            variants = get_literal_variants(var_toks)
            assert all([len(variant) <= self.max_len for variant in variants]), variants
            for variant in variants:
                for prefix_len in range(1, len(variant) + 1):
                    if toks_eq(
                        toks[partial.end : partial.end + prefix_len],
                        variant[:prefix_len],
                    ):
                        res.append(
                            PartialMatch(
                                name=self.name,
                                start=partial.end,
                                end=partial.end + prefix_len,
                                defns=partial.defns,
                                data=self.var_len_and_prefix_idx(
                                    len(variant), prefix_len
                                ),
                            )
                        )
        return res

    @property
    def pyregex(self):
        return r".{0," + str(self.max_len) + r"}"


import regex as re


class TokRegex(nn.Module):
    def __init__(self, pattern, search=False):
        super().__init__()
        self.name = f"TokRegex:{pattern}:{randstr()}"
        self.pattern = pattern
        self.search = search

    def matches(self, toks, partial, reversed):
        if partial.end == len(toks):
            return []

        tok = toks[partial.end]

        if (self.search and re.search(self.pattern, tok)) or re.fullmatch(
            self.pattern, tok
        ):
            return [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end + 1,
                    defns=partial.defns,
                    data=None,
                )
            ]
        else:
            return []

class FlexRegex(nn.Module):
    def __init__(self, pattern, search=False):
        super().__init__()
        self.name = f"FlexRegex:{pattern}:{randstr()}"
        self.pattern = pattern
        self.search = search
        self.spacing = Embed(2)
        self.capitalization = Embed(6)
        self.mixer = Mixer(2, linear=True, bilinear=True)

        

    def matches(self, toks, partial, reversed):
        if partial.end == len(toks):
            return []


        tok = toks[partial.end]

        stripped_tok = tok.strip()
        normalized_tok = stripped_tok.lower()
        

        if (self.search and re.search(self.pattern, normalized_tok)) or re.fullmatch(
            self.pattern, normalized_tok
        ):
            capitalization = 0 if len(stripped_tok) == 0 else\
                  1 if stripped_tok.islower() else\
                  2 if stripped_tok.isupper() else\
                  3 if (stripped_tok[0].isupper() and len(stripped_tok) > 1 and stripped_tok[1:].islower()) else\
                  4 if stripped_tok[0].isupper() else\
                  5
            return [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end + 1,
                    defns=partial.defns,
                    data=[self.spacing(int(tok[0] == ' ')), self.capitalization(capitalization)],
                )
            ]
        else:
            return []


class TokRegexSet(nn.Module):
    def __init__(self, pattern, search=False):
        super().__init__()
        self.name = f"TokRegexSet:{pattern}:{randstr()}"
        self.pattern = pattern
        self.search = search

        if search is True:
            self.toks = {
                tok for tok in tokre.get_all_toks() if re.search(self.pattern, tok)
            }
        else:
            self.toks = {
                tok for tok in tokre.get_all_toks() if re.fullmatch(self.pattern, tok)
            }

    def matches(self, toks, partial, reversed):
        if partial.end == len(toks) and not reversed:
            return []

        if partial.end == len(toks) and reversed:
            return []

        tok = toks[partial.end]

        if tok in self.toks:
            return [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end + 1,
                    defns=partial.defns,
                    data=None,
                )
            ]
        else:
            return []


class Prefix(nn.Module):
    def __init__(self, child_module, max_len=10):
        super().__init__()
        self.name = f"Prefix:{randstr()}"
        self.child_module = child_module
        self.max_len = max_len
        self.match_len_and_prefix_len = Embed((max_len, max_len))
        self.mixer = Mixer(2, linear=True, bilinear=True)

    def matches(self, toks, partial, reversed):
        matches = self.child_module.matches(toks, partial, reversed)
        res = []
        for match in matches:
            for prefix_end in range(match.start+1, match.end+1):
                res.append(
                    PartialMatch(
                        name=self.name,
                        start=match.start,
                        end=prefix_end,
                        defns=match.defns,
                        data=[match, self.match_len_and_prefix_len(match.end-match.start, prefix_end - match.start)]
                    )
                )
        return res


import tokre
import json

import torch
from tokre.core.modules import PartialMatch


class TrieNode:
    def __init__(self):
        self.children = {}
        self.is_end = False
        self.value = None


class Trie:
    def __init__(self, literals, values=None):
        self.root = TrieNode()
        if values is None:
            values = literals
        for literal, value in zip(literals, values):
            self.insert(literal, value)

    def insert(self, literal, value):
        node = self.root
        for tok_string in literal:
            if tok_string not in node.children:
                node.children[tok_string] = TrieNode()
            node = node.children[tok_string]
        node.is_end = True
        node.value = value

    def prefixes(self, token_strs):
        result = []
        node = self.root
        for i, tok_string in enumerate(token_strs):
            if tok_string not in node.children:
                break
            node = node.children[tok_string]
            if node.is_end:
                result.append(node.value)
        return result


class LiteralSet(nn.Module):
    def __init__(self, literal_name):
        super().__init__()
        self.name = f"Literalset:{literal_name}:{randstr()}"

        with open(tokre.get_workspace() / (literal_name + ".json")) as f:
            data = json.load(f)
        assert isinstance(data, dict)
        assert "literal_toks" in data

        self.literal_set = [tuple(it) for it in data["literal_toks"]]
        self.trie = Trie(self.literal_set)

        self.reversed_trie = Trie([toks[::-1] for toks in self.literal_set])

        self.literal_idx = Embed(len(self.literal_set))

        self.mixer = Mixer(1, linear=True)

    def matches(self, toks, partial, reversed):
        trie = self.trie if reversed is False else self.reversed_trie

        res = []
        forward_toks = toks[partial.end :]

        matching_prefixes = trie.prefixes(forward_toks)
        for prefix in matching_prefixes:
            prefix = prefix if reversed is False else prefix[::-1]
            res.append(
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end + len(prefix),
                    defns=partial.defns,
                    data=[self.literal_idx(self.literal_set.index(prefix))],
                )
            )

        return res
    
from frozendict import frozendict

class Redefine(nn.Module):
    def __init__(self, var_name: str, new_var_name: str, ):
        self.name = f'Redefine:{randstr()}'
        self.var_name = var_name
        self.new_var_name = new_var_name

    def matches(self, toks, partial, reversed):
        if self.var_name in partial.defns:
            new_defns = {k: v for (k, v) in partial.defns.items() if k != self.var_name}
            new_defns[self.new_var_name] = partial.defns[self.var_name]
            return [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end,
                    defns = frozendict(new_defns),
                    data=None
                )
            ]
        else:
            return [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end,
                    defns = partial.defns,
                    data=None
                )
            ]

        



DEFINED_MACROS = {
    "var_variant_prefix": VarVariantPrefix,
    "re": TokRegex,
    "re_tok_set": TokRegexSet,
    "literal_set": LiteralSet,
    "literals": LiteralSet,
    "prefix": Prefix,
    "var_variant": VarVariant,
    "BEGIN": BEGIN,
    "redefine": Redefine,
    "pos": AbsPos,
    'flex': FlexRegex
}



import lark
from lark import Lark, Transformer, Tree, Discard
import regex as re

grammar = r"""
    %import common.INT

    SPACE: " "
    SPACES: SPACE+
    pad: SPACES

    NUMBER: INT
    WILDCARD: /[.]/
    ESCAPED: /\\[*+|.()\[\]{}?=<>\\0-9n]/

    REPEAT_POSTFIX.1: /[+*?]/ | "{" SPACES? ( NUMBER? SPACES? "," SPACES? NUMBER | NUMBER SPACES? "," NUMBER? | NUMBER) SPACES? "}"

    ?non_special_chars: /[^*+|.()\[\]{}?=<>\\0-9]/
    char_seq: (non_special_chars | ESCAPED | NUMBER | SPACES)+

    pos_lookbehind: "(?<=" pattern ")"
    pos_lookahead: "(?=" pattern ")"
    neg_lookbehind: "(?<!" pattern ")"
    neg_lookahead: "(?!" pattern ")"

    ?lookaround: (pos_lookbehind | pos_lookahead | neg_lookbehind | neg_lookahead)
    
    identifier_chars: /[a-zA-Z_]/
    identifier: identifier_chars+

    macro_string_arg: "`" (/[^`]/)* "`"
    bool: /False|True/
    ?macro_arg: (group_block | macro_string_arg | NUMBER | bool)
    macro_kwarg: identifier "=" macro_arg
    macro: "[" pad? identifier (pad (macro_arg| macro_kwarg))* pad? "]"

    var_defn: "[" pad? identifier pad? "=" pattern "]"

    #  <group_block>{..}
    repeated_group: group_block REPEAT_POSTFIX

    # clearly delineated block
    # (...) | (..|..) | [...] | . | [..=..] | ..among others..
    ?group_block: "(" (or_pattern | phrase) ")" | macro | var_defn | lookaround | repeated_group | WILDCARD
    
    or_pattern: phrase ("|" phrase)+

    # ( group | group{...} | seq of chars)+
    ?phrase: (group_block  | char_seq )+

    ?pattern: phrase | or_pattern

    NEWLINE: "\n"

    # line_defn: identifier pad? "=" " (Lily)"
    
    ?lines: (NEWLINE* (pattern)) (NEWLINE+ (pattern))* NEWLINE*

    ?start   : lines
"""

special_chars_wo_newline = r"\^$|&*+{}()<>?[].=;" + "".join([str(i) for i in range(10)])

special_tokre_chars = special_chars_wo_newline + "\n"


def escape(s):
    """
    Escapes special characters in a string for use in tokre patterns.

    Args:
    s (str): The input string to escape.

    Returns:
    str: The escaped string.
    """
    global special_chars_wo_newline
    s = "".join("\\" + c if c in special_chars_wo_newline else c for c in s)
    s = s.replace("\n", "\\n")
    return s


parser = Lark(grammar, parser="lalr")


class SimplifyTree(Transformer):
    def __init__(self):
        super().__init__()
        self.line_defns = {}
        self.var_defns = {}

    def pad(self, items):
        return Discard

    def NEWLINE(self, items):
        return Discard

    def repeated_group(self, items):
        assert len(items) == 2, items
        group, repeat_str = items[0], items[1]

        if repeat_str == "+":
            min, max = 1, float("inf")
        elif repeat_str == "*":
            min, max = 0, float("inf")
        elif repeat_str == "?":
            min, max = 0, 1
        else:
            assert repeat_str[0] == r"{" and repeat_str[-1] == r"}"

            # remove curly brackets
            repeat_str = repeat_str[1:-1]

            if "," in repeat_str:
                min, max = repeat_str.split(",")
                min, max = min.strip(), max.strip()
                min = 0 if min == "" else int(min)
                max = float("inf") if max == "" else int(max)
            else:
                num = int(repeat_str.strip())
                min, max = num, num

        assert not (min > max), f"min > max in parsed repeat: {min=} {max=}"

        # return Repeat(child_matcher=group, min=min, max=max)
        return Tree("repeat", [group, min, max])

    def char_seq(self, children):
        s = ""
        for child in children:
            assert isinstance(child, lark.lexer.Token) or\
                isinstance(child, int),\
                    child
            child_str = str(child)
    
            if len(child_str) == 2 and child_str[0] == "\\":
                if child_str[1] == 'n':
                    child_str = '\n'
                else:
                    child_str = child_str[-1]

            s += child_str
        if len(s.replace('\n', '↵').strip()) == 0:
            return Discard

        s = s.rstrip(' ')
        if s[0] == " ":
            s = " " + s.lstrip()
        return Tree("string", [s])

    def identifier_chars(self, chars):
        s = "".join([str(c) for c in chars])
        return s

    def identifier(self, chars):
        return "".join(chars)

    def macro(self, args):
        args = [
            arg
            for arg in args
            if not (isinstance(arg, lark.lexer.Token) and arg.type == "SPACES")
        ]
        macro_name = args[0]
        assert isinstance(macro_name, str)

        if macro_name in self.var_defns:
            return Tree("var_ref", [macro_name])
        elif macro_name in self.line_defns:
            return self.line_defns[macro_name]
        else:
            return Tree("macro", [macro_name, *args[1:]])

    def pattern(self, args):
        return Tree("pat", args)

    def var_defn(self, children):
        children = [
            child
            for child in children
            if not (isinstance(child, lark.lexer.Token) and child.type == "SPACES")
        ]
        assert len(children) == 2
        var_name, child_tree = children

        assert isinstance(var_name, str)
        self.var_defns[var_name] = child_tree

        child_tree = children[1]

        return Tree("var_defn", [var_name, child_tree])

    def pos_lookahead(self, children):
        assert len(children) == 1
        child_tree = children[0]

        # [child, is_backward, is_neg]
        return Tree("lookaround", [child_tree, False, False])

    def pos_lookbehind(self, children):
        assert len(children) == 1
        child_tree = children[0]

        # [child, is_backward, is_neg]
        return Tree("lookaround", [child_tree, True, False])

    def neg_lookahead(self, children):
        assert len(children) == 1
        child_tree = children[0]

        # [child, is_backward, is_neg]
        return Tree("lookaround", [child_tree, False, True])

    def neg_lookbehind(self, children):
        assert len(children) == 1
        child_tree = children[0]

        # [child, is_backward, is_neg]
        return Tree("lookaround", [child_tree, True, True])

    def macro_string_arg(self, children):
        return "".join([str(ch) for ch in children])

    def or_pattern(self, children):
        return Tree("or_pattern", children)

    def line_defn(self, children):
        assert len(children) == 2
        defn_name, child_tree = children
        self.line_defns[defn_name] = child_tree
        return Discard

    def lines(self, lines):
        if len(lines) > 1:
            return Tree("phrase", lines)
        else:
            assert (
                len(lines) == 1
            ), "Don't see at least at least one line of code; did you feed in only line definitions?"
            return lines[0]

    def phrase(self, children):
        if len(children) == 1:
            assert isinstance(children[0], Tree), "non-tree single child in phrase"
            return children[0]
        else:
            return Tree("phrase", children)

    def bool(self, children):
        assert len(children) == 1
        val = bool(str(children[0]))
        return val

    def NUMBER(self, num_token):

        val = int(str(num_token))
        return val

    def macro_kwarg(self, children):
        assert len(children) == 2
        assert isinstance(children[0], str)

        return {children[0]: children[1]}

    def WILDCARD(self, chilren):
        """
        need to convert to tree before InsertModules
        or an error gets thrown if a tokre-code line is only a wildcard.
        Can't apply a Transformer to a Token object.
        """
        return Tree("wildcard", [])


def parse_line(line: str):
    assert "\n" not in line
    # [non-newline space]* [name=...] [non-newline space]*
    match = re.match(r"[^\S\r\n]*(?P<name>[a-zA-Z_]+)[^\S\r\n]*=", line)
    if match:
        assert "name" in match.groupdict()
        defn_name = match.groupdict()["name"]
        pattern = line[match.end() :]
        return Tree("line_defn", [defn_name, parser.parse(pattern)])
    else:
        return parser.parse(line)


def remove_comments(code):
    # Remove inline comments
    code = re.sub(r"(?<!\\)#.*", "", code)
    # # Remove multiline comments
    # code = re.sub(r'\'\'\'.*?\'\'\'', '', code, flags=re.DOTALL)
    # code = re.sub(r'\"\"\".*?\"\"\"', '', code, flags=re.DOTALL)
    return code


def parse(s):
    s = remove_comments(s)
    parsed_lines = [parse_line(line) for line in s.split("\n") if len(line.strip()) > 0]
    tree = Tree("lines", parsed_lines)
    tree = SimplifyTree().transform(tree)
    return tree



import hashlib
import json
from queue import Queue
from threading import Thread

from tqdm import tqdm


def get_vocab_size(tokenizer):
    """
    Get the size of the vocabulary from the tokenizer.

    Args:
        tokenizer: The tokenizer object which should have a `vocab_size` or `n_vocab` attribute.

    Returns:
        int: The size of the vocabulary.
    """
    if hasattr(tokenizer, "vocab_size"):
        return tokenizer.vocab_size
    elif hasattr(tokenizer, "n_vocab"):
        return tokenizer.n_vocab
    else:
        raise AttributeError(
            "The tokenizer must have either 'vocab_size' or 'n_vocab' attribute."
        )


def hash_tokenizer(tokenizer):
    """
    Hash a tokenizer by deterministically hashing its vocabulary.
    Requires the tokenizer object to have:
        - a `vocab_size` or `n_vocab` attribute.
        - a `decode` method that can convert token IDs to strings.
    """
    assert hasattr(
        tokenizer, "decode"
    ), "Tokenizer must have a `decode` method for `hash_tokenizer` to work."
    assert hasattr(tokenizer, "vocab_size") or hasattr(
        tokenizer, "n_vocab"
    ), "Tokenizer must have either a `vocab_size` or `n_vocab` attribute for `hash_tokenizer` to work."

    if hasattr(tokenizer, "vocab_size"):
        assert isinstance(tokenizer.vocab_size, int)
        vocab_size = tokenizer.vocab_size  # attribute for HF tokenizers
    elif hasattr(tokenizer, "n_vocab"):
        assert isinstance(tokenizer.n_vocab, int)
        vocab_size = tokenizer.n_vocab  # attribute for tiktoken tokenizers
    else:
        raise ValueError(
            "Tokenizer must have either n_vocab or vocab_size attribute for `hash_tokenizer` to work. If your tokenizer isn't custom, are you using a HF or tiktoken tokenizer?"
        )

    # Could be sped up for >3M token tokenizers by sampling random tokens for tokens_str using a fixed random seed.
    # python hash fn is nondeterministic between sessions (!) so we use md5 instead.
    tokens_str = tokenizer.decode(list(range(vocab_size)))
    return hashlib.md5(tokens_str.encode()).hexdigest()


def threaded_map(
    fn,
    inputs,
    pbar_desc=None,
    args=None,
    kwargs=None,
    n_threads=20,
    input_is_kwargs=False,
):
    args = [] if args is None else args
    kwargs = {} if kwargs is None else kwargs

    pbar_desc = f"threaded_map" if pbar_desc is None else pbar_desc

    pbar = tqdm(total=len(inputs), desc=pbar_desc)
    input_queue = Queue()
    for idx, item in enumerate(inputs):
        input_queue.put((idx, item))
    results_queue = Queue()

    def worker():
        while not input_queue.empty():
            idx, item = input_queue.get()
            if input_is_kwargs:
                result = fn(*args, **item, **kwargs)
            else:
                result = fn(item, *args, **kwargs)
            results_queue.put((idx, result))
            input_queue.task_done()
            pbar.update(1)

    for _ in range(n_threads):
        thread = Thread(target=worker)
        thread.daemon = True
        thread.start()

    input_queue.join()

    results = []
    while not results_queue.empty():
        results.append(results_queue.get())

    return [result for idx, result in sorted(results, key=lambda x: x[0])]


def format_nest(nest, **kwargs):
    """
    nest is a nest of lists, dictionaries, and strings.
    returns the same nest but with the strings formatted using provided kwargs.
    """
    if isinstance(nest, list):
        return [format_nest(item, **kwargs) for item in nest]
    elif isinstance(nest, dict):
        return {
            format_nest(k, **kwargs): format_nest(v, **kwargs) for k, v in nest.items()
        }
    elif isinstance(nest, str):
        return nest.format(**kwargs)


def assert_snake_case(s):
    assert "-" not in s, "snake_case please"
    assert " " not in s, "snake_case please"
    assert s.lower() == s, "snake_case please"


def save_dict(d, fname):
    s = "{"
    for k, v in d.items():
        s += "\n  " + f'"{k}": ' + json.dumps(v) + ","
    s = s[:-1]
    s += "\n}\n"
    with open(fname, "w") as f:
        f.write(s)



import numpy as np
import torch
from functools import lru_cache
import unicodedata
from tokre.core.parsing import special_tokre_chars
from typing import Iterable
import tokre


def is_valid_char(code_point):
    try:
        char = chr(code_point)
        # If the character is printable, has a name, and is not a special char, it is valid.
        if (
            char.isprintable()
            and unicodedata.name(char, None)
            and char not in special_tokre_chars
        ):
            return True
    except (UnicodeEncodeError, UnicodeDecodeError):
        pass
    return False


@lru_cache
def valid_unicode_characters():
    valid_unicode_characters = []
    for code_point in range(0x110000):  # Unicode code points range from 0 to 0x10FFFF
        if is_valid_char(code_point):
            valid_unicode_characters.append(chr(code_point))
    return valid_unicode_characters


VALID_CHARS = np.array(valid_unicode_characters())


def pyregex_literal(toks):
    if (
        isinstance(toks, Iterable)
        and not isinstance(toks, str)
        and isinstance(toks[0], str)
    ):
        encoded_toks = [tokre.encode(tok) for tok in toks]
        assert all([len(tok_ids) == 1 for tok_ids in encoded_toks])
        toks = [tok_ids[0] for tok_ids in encoded_toks]

    if isinstance(toks, np.ndarray) or isinstance(toks, list):
        toks = torch.tensor(toks)
    elif isinstance(toks, torch.Tensor):
        toks = toks.cpu()
    return "".join(VALID_CHARS[toks])



from pathlib import Path
from typing import Optional

import tokre
from tokre.utils import hash_tokenizer


def setup(
    tokenizer, workspace: Optional[str] = None, openai_api_key: Optional[str] = None
):
    """
    Sets up tokre.

    Before being able to use tokre, you must provide a tokenizer class that implements `encode` and `decode`.

    If workspace is provided, it's used as a directory to load and store scripts.
    If the workspace doesn't exist in the current directory, a directory is created.

    """
    assert tokenizer is not None, "`tokenizer` needs to be provided."
    if workspace is not None:
        assert workspace != "", "`workspace` can't be an empty string."

    if openai_api_key is not None:
        tokre._openai_api_key = openai_api_key

    if workspace is None:
        if tokre._workspace is not None:
            assert (
                False
            ), "Cannot set only `tokenizer` using tokre.setup when `workspace` is already set. Please re-run tokre.setup with both a `tokenizer` and `workspace` argument."
        tokre._tokenizer = tokenizer

    else:
        # (if workspace is not None)

        ws_path = Path(workspace)

        # If the workspace doesn't exist, create it with .tokenizer_hash
        if not ws_path.is_dir():
            ws_path.mkdir(parents=True)

            # Write a hash of the tokenizer to .tokenizer_hash
            with open(ws_path / ".tokenizer_hash", "w") as f:
                f.write(hash_tokenizer(tokenizer))

        # If the workspace exists, make sure the tokenizer matches
        else:
            tokenizer_hash_path = ws_path / ".tokenizer_hash"

            # make sure .tokenizer_hash is inside the workspace
            assert tokenizer_hash_path.exists(), (
                f".tokenizer_hash wasn't found inside {ws_path}.\n"
                "Pre-existing workspace should have a tokenizer hash stored. Aborting workspace setup."
            )

            # make sure .tokenizer_hash matches the provided tokenizer's hash
            assert tokenizer_hash_path.read_text() == hash_tokenizer(tokenizer), (
                f"Hash of the provided tokenizer doesn't match the hash stored in {tokenizer_hash_path}\n",
                f"Aborting workspace setup.",
            )

        ws_path = ws_path.resolve()  # resolve makes it an absolute path

        tokre._workspace = ws_path
        tokre._tokenizer = tokenizer



from __future__ import annotations
import torch
import torch.nn as nn

from typing import Union, TypeAlias, Optional
from dataclasses import dataclass
from frozendict import frozendict


def randstr(length=6):
    import random
    import string

    return "".join(random.choices(string.ascii_uppercase, k=length))


@dataclass
class EmbedData:
    name: str
    data: Union[int, tuple[int]]


class Embed(nn.Module):
    def __init__(
        self, embed_shape: Union[list[int], tuple[int], int], name: Optional[str] = None
    ):
        super().__init__()

        self.name = "Embed:" + randstr() if name is None else name

        if isinstance(embed_shape, int):
            embed_shape = (embed_shape,)

        embed = torch.ones(embed_shape)

        self.embed_shape = embed_shape
        self.embed = nn.Parameter(embed)

    def __call__(self, *indices):
        for i, index in enumerate(indices):
            assert index >= 0, "index must be >= 0"
            assert index < self.embed_shape[i], "index too large"

        if len(indices) == 1:
            indices = indices[0]

        return EmbedData(name=self.name, data=indices)

    def pred(self, embed_data: EmbedData):
        indices = tuple(embed_data.data)
        assert all([isinstance(i, int) for i in indices])
        assert len(indices) == len(self.embed_shape)

        assert all(
            [
                (idx >= 0 and idx < embed_max)
                for (idx, embed_max) in zip(indices, self.embed_shape)
            ]
        )

        return self.embed[tuple(embed_data.data)]

    def __repr__(self):
        return f"Embed{self.embed_shape}"


class Mixer(nn.Module):
    def __init__(self, d_module: int, bilinear=False, linear=False):
        super().__init__()
        assert bilinear is True or linear is True  # or d_module == 1
        assert d_module > 0

        self.d_module = d_module

        self.bilinear = bilinear
        self.linear = linear

        if self.bilinear:
            self.bilinear_pre_bias = nn.Parameter(
                torch.zeros(
                    d_module + 1,
                )
            )
            self.bilinear_param = nn.Parameter(torch.zeros(d_module + 1, d_module + 1))

        if self.linear:
            self.linear_pre_bias = nn.Parameter(
                torch.zeros(
                    d_module + 1,
                )
            )
            self.linear_param = nn.Parameter(torch.ones(d_module + 1) / d_module)

        self.bias = nn.Parameter(torch.zeros(1)[0])

    def device(self):
        if hasattr(self, "bilinear_param"):
            return self.bilinear_param.device
        elif hasattr(self, "linear_param"):
            return self.linear_param.device
        else:
            raise ValueError(
                "Mixer object has neither self.bilinear_param or self.linear_param, which isn't expected."
            )

    def dtype(self):
        if hasattr(self, "bilinear_param"):
            return self.bilinear_param.dtype
        elif hasattr(self, "linear_param"):
            return self.linear_param.dtype
        else:
            raise ValueError(
                "Mixer object has neither self.bilinear_param or self.linear_param, which isn't expected."
            )

    def forward(self, x):
        D = x.shape[0]
        y = self.bias
        if self.bilinear:
            pre_bilinear = x  # + self.bilinear_pre_bias[:D]
            y = y + torch.einsum(
                "i, ij, j", pre_bilinear, self.bilinear_param[:D, :D], pre_bilinear
            )
        if self.linear:
            pre_linear = x  # + self.linear_pre_bias[:D]
            y = y + self.linear_param[:D] @ pre_linear
        return y

    # def forward(self, x):
    #     assert len(x.shape) == 2, "input must be batched vectors"

    #     if self.linear and not self.bilinear:
    #         return x @ self.linear_param
    #     elif self.bilinear and not self.linear:
    #         return torch.einsum("mn, bn, bm -> b", self.bilinear_param, x, x)
    #     elif self.linear and self.bilinear:
    #         return (x @ self.linear_param) + torch.einsum(
    #             "mn, bn, bm -> b", self.bilinear_param, x, x
    #         )
    #     else:
    #         assert self.d_module == 1
    #         return x[..., 0]

    # def pred(self, data: list[PredData]):

    def __repr__(self):
        return f"Mixer({self.d_module}, bilinear={self.bilinear}, linear={self.linear})"


PredData = Union["PartialMatch", EmbedData, None, list["PredData"]]


@dataclass
class PartialMatch:
    name: str
    start: int
    end: int
    defns: frozendict
    data: PredData

    def __len__(self):
        return self.end - self.start


def is_pred_data(obj):
    return (
        isinstance(obj, (PartialMatch, EmbedData, list))
        or obj is None
        or (isinstance(obj, list) and all(is_pred_data(item) for item in obj))
    )


def batched_extend_matches(
    toks, partial_matches: list[PartialMatch], child_matcher, reversed
):
    new_partials = []

    for partial in partial_matches:
        assert isinstance(
            partial.data, list
        ), f"Provided partial match group_data must be represented as a list: {partial.data=}"
        matcher_results = child_matcher.matches(toks, partial, reversed=reversed)

        for match_extension in matcher_results:
            extended_match = PartialMatch(
                name=partial.name,
                start=partial.start,
                end=match_extension.end,
                defns=match_extension.defns,
                data=partial.data + [match_extension],
            )
            new_partials.append(extended_match)
    return new_partials


# def pyregex_literal(toks):
#     # [STUB]
#     return "".join(toks)
from tokre.core.pyregex import pyregex_literal


def toks_eq(toks_a: list[str], toks_b: list[str]):
    return (len(toks_a) == len(toks_b)) and all(
        [a == b for (a, b) in zip(toks_a, toks_b)]
    )


class Toks(nn.Module):
    def __init__(self, toks: list[str], name: Optional[str] = None):
        super().__init__()
        self.name = "Toks:" + randstr() if name is None else name
        self.toks = toks

    def matches(self, toks, partial, reversed: bool):
        match_toks = self.toks if reversed is False else self.toks[::-1]
        if toks_eq(toks[partial.end : partial.end + len(match_toks)], match_toks):
            match = PartialMatch(
                name=self.name,
                start=partial.end,
                end=partial.end + len(match_toks),
                defns=partial.defns,
                data=None,
            )
            return [match]
        else:
            return []

    @property
    def pyregex(self):
        return pyregex_literal(self.toks)

    def __repr__(self):
        return f"Toks({self.toks})"


Inf = float("inf")


class Repeat(nn.Module):
    def __init__(self, child_matcher, min: int, max: Union[int, Inf], name=None):
        super().__init__()
        self.name = "Repeat:" + randstr() if name is None else name
        self.child_matcher = child_matcher
        self.min = min
        self.max = max

        self.n_repeats = Embed(128, "n_repeats")

        if max == Inf:
            self.d_mixer = 1
        else:
            assert isinstance(max, int)
            self.n_repeats = Embed(max + 1)
            self.d_mixer = max + 1
            if self.d_mixer > 1:
                self.mixer = Mixer(self.d_mixer, linear=True)

    def matches(self, toks, partial, reversed):
        starting_partial = PartialMatch(
            name=self.name,
            start=partial.end,
            end=partial.end,
            defns=partial.defns,
            data=[],
        )

        res = []
        new_partials = [starting_partial]
        if self.min == 0:
            res.extend(new_partials)

        repeat_idx = 1
        while repeat_idx <= self.max and new_partials:
            new_partials = batched_extend_matches(
                toks, new_partials, self.child_matcher, reversed=reversed
            )
            if repeat_idx >= self.min:
                res.extend(new_partials)
            repeat_idx += 1

        # Special data if self.max == Inf
        if self.max == Inf:
            for partial in res:
                partial.data = self.n_repeats(len(partial.data))

        return res

    def extra_repr(self):
        return f"""(min): {self.min}
(max): {self.max}"""

    @property
    def pyregex(self):
        pyregex_max = "" if self.max == Inf else str(self.max)
        return (
            f"({self.child_matcher.pyregex})"
            + r"{"
            + str(self.min)
            + ","
            + pyregex_max
            + r"}"
        )


class Phrase(nn.Module):
    def __init__(self, matchers, name=None):
        super().__init__()
        self.name = "Phrase:" + randstr() if name is None else name
        self.matchers = nn.ModuleList(matchers)

        self.mixer = Mixer(len(self.matchers), linear=True, bilinear=True)

    def matches(self, toks, partial, reversed: bool):
        starting_partial = PartialMatch(
            name=self.name,
            start=partial.end,
            end=partial.end,
            defns=partial.defns,
            data=[],
        )

        partials = [starting_partial]

        for matcher in self.matchers:
            partials = batched_extend_matches(
                toks, partials, matcher, reversed=reversed
            )
        return partials

    @property
    def pyregex(self):
        return f"({''.join([child.pyregex for child in self.matchers])})"


class Wildcard(nn.Module):
    def __init__(self, name=None):
        super().__init__()
        self.name = "Wildcard:" + randstr() if name is None else name

    def matches(self, toks, partial, reversed: bool):
        if partial.end < len(toks):
            return [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end + 1,
                    defns=partial.defns,
                    data=None
                )
            ]
        else:
            return []

    @property
    def pyregex(self):
        return f"."


class OrGroup(nn.Module):
    def __init__(self, matchers, name=None):
        super().__init__()
        self.name = "OrGroup:" + randstr() if name is None else name
        self.branches = nn.ModuleList(matchers)

        self.which_branch = Embed(len(self.branches))

        self.mixer = Mixer(2, linear=True)

    def matches(self, toks, partial, reversed=False):
        res = []
        for branch_idx, branch in enumerate(self.branches):
            for match in branch.matches(toks, partial, reversed=reversed):
                res.append(
                    PartialMatch(
                        name=self.name,
                        start=partial.end,
                        end=match.end,
                        defns=match.defns,
                        data=[self.which_branch(branch_idx), match],
                    )
                )
        return res

    @property
    def pyregex(self):
        return f"({'|'.join([child.pyregex for child in self.branches])})"


class VarDefn(nn.Module):
    def __init__(self, var_name, child_matcher, name=None):
        super().__init__()
        self.name = f"VarDefn:{var_name}:{randstr()}" if name is None else name
        self.var_name = var_name
        self.child_matcher = child_matcher

    def matches(self, toks, partial, reversed=False):
        child_matches = self.child_matcher.matches(toks, partial, reversed=reversed)

        res = []
        for match in child_matches:
            res.append(
                PartialMatch(
                    name=self.name,
                    start=match.start,
                    end=match.end,
                    defns=match.defns | {self.var_name: toks[match.start : match.end]},
                    data=match,
                )
            )

        return res

    @property
    def pyregex(self):
        f"(?P<{self.var_name}>{self.child_matcher.pyregex})"

    def extra_repr(self):
        return f"(var_name): '{self.var_name}'"


class VarRef(nn.Module):
    def __init__(self, var_name, name=None):
        super().__init__()
        self.name = f"VarRef({var_name}):{randstr()}" if name is None else name
        self.var_name = var_name

    def matches(self, toks, partial, reversed):
        # unaffected by reversed

        if self.var_name not in partial.defns:
            return []
        var_defn = partial.defns[self.var_name]

        if toks_eq(toks[partial.end : partial.end + len(var_defn)], var_defn):
            return [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end + len(var_defn),
                    defns=partial.defns,
                    data=None,
                )
            ]
        else:
            return []

    def extra_repr(self):
        return f"'{self.var_name}'"


"""
 # [child, is_backward, is_neg]
Tree("lookaround", [child_tree, True, False])
"""


# def reverse_transformation(toks, partial):
#     toks = toks[::-1]
#     partial = PartialMatch(
#         name=partial.name,
#         start=len(toks)-partial.start
#         end=len(toks)-partial.end
#     )
# class Lookaround(nn.Module):
#     def __init__(self, child_module, is_backward: bool, is_neg: bool, name=None):
#         super().__init__()
#         self.name = f'Lookaround:{randstr()}' if name is None else name
#         self.child_module = child_module
#         self.is_backward = is_backward
#         self.is_neg = is_neg

#     def matches(self, toks, partial, reversed):
#         toks = toks if not reversed else toks[::-1]
#         cur_idx = partial.end if not reversed else len(toks)-partial.end-1
#         # altered_partial = Partial(
#         #     name=self.name,
#         #     start=
#         # )
#         matches = self.child_module.matches(toks, partial, reversed=(not reversed if self.is_backward else reversed))
#         if self.is_neg:
#             if matches:
#                 return []
#             else:
#                 return [partial]
#         else:
#             matches = [
#                 PartialMatch(
#                     name=self.name,
#                     start=cur_idx,
#                     end=cur_idx,
#                     defns=match.defns,
#                     data=match
#                 )
#                 for match in matches
#             ]
#         return matches


class Lookahead(nn.Module):
    def __init__(self, child_module, is_neg: bool, name=None):
        super().__init__()
        self.name = f"Lookaround:{randstr()}" if name is None else name
        self.child_module = child_module
        self.is_neg = is_neg

    def matches(self, toks, partial, reversed):

        matches = self.child_module.matches(toks, partial, reversed=False)
        if self.is_neg:
            if matches:
                return []
            else:
                return [
                    PartialMatch(
                        name=self.name,
                        start=partial.end,
                        end=partial.end,
                        defns=partial.defns,
                        data=None,
                    )
                ]
        else:
            matches = [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end,
                    defns=match.defns,
                    data=match,
                )
                for match in matches
            ]
        return matches

    def extra_repr(self):
        return f"(is_neg): {self.is_neg}"


class Lookbehind(nn.Module):
    def __init__(self, child_module, is_neg: bool, name=None):
        super().__init__()
        self.name = f"Lookaround:{randstr()}" if name is None else name
        self.child_module = child_module
        self.is_neg = is_neg

    def matches(self, toks, partial, reversed):
        reversed_toks = toks[::-1]
        reversed_partial = PartialMatch(
            name=partial.name,
            start=len(toks) - partial.end,
            end=len(toks) - partial.end,
            defns=frozendict({k: v[::-1] for k, v in partial.defns.items()}),
            data=partial.data,
        )

        matches = self.child_module.matches(
            reversed_toks, reversed_partial, reversed=True
        )
        if self.is_neg:
            if len(matches) > 0:
                return []
            else:
                return [
                    PartialMatch(
                        name=self.name,
                        start=partial.end,
                        end=partial.end,
                        defns=partial.defns,
                        data=None,
                    )
                ]
        else:
            matches = [
                PartialMatch(
                    name=self.name,
                    start=partial.end,
                    end=partial.end,
                    defns={k: v[::-1] for k, v in match.defns.items()},
                    data=match,
                )
                for match in matches
            ]
        return matches

    def extra_repr(self):
        return f"(is_neg): {self.is_neg}"


class LearnedConst(nn.Module):
    def __init__(self, child_module, name=None):
        super().__init__()
        self.name = f"LearnedConst:{randstr()}" if name is None else name
        self.child_module = child_module
        self.bias = Embed(1)

    def matches(self, toks, partial, reversed):
        matches = self.child_module.matches(toks, partial, reversed)
        matches = [
            PartialMatch(
                name=self.name,
                start=match.start,
                end=match.end,
                defns=match.defns,
                data=self.bias(0),
            )
            for match in matches
        ]
        return matches

    @property
    def pyregex(self):
        return self.child_module.pyregex



import itertools
import json
import os
import random
from functools import lru_cache
from typing import Optional

import numpy as np

import tokre
from tokre.labelling.query_openai import query_openai
from tokre.utils import assert_snake_case, format_nest, get_vocab_size, threaded_map

from tokre.utils import save_dict


@lru_cache()
def get_all_toks():
    tok_ids = np.arange(get_vocab_size(tokre._tokenizer))
    toks = [tokre.decode([tok_id]) for tok_id in tok_ids]
    return toks


def extract_stripped_toks(toks):
    return list(set([t.strip() for t in toks]))


def extract_space_toks(toks):
    # returns tokens that start with a space
    toks = [t for t in toks if t.strip() != ""]
    return list(set([t for t in toks if t[0] == " "]))


def extract_no_space_toks(toks):
    toks = [t for t in toks if t.strip() != ""]
    return list(set([t for t in toks if t[0] != " "]))


def extract_capitalized_toks(toks):
    # returns tokens that are capitalized
    toks = [t for t in toks if t.strip() != ""]
    return list(set([t for t in toks if t.strip()[0].isupper()]))


def extract_not_capitalized_toks(toks):
    # returns tokens that are not capitalized
    toks = [t for t in toks if t.strip() != ""]
    return list(set([t for t in toks if t.strip()[0].islower()]))


prompt_template = """Given the list of strings below, identify which strings match the following description:
{description}
To log the strings that match the description, call log_{label}_strings([...identified strings]). Copy strings exactly as they appear.

Here is the list of strings for you to process: {batch}"""

functions_template = [
    {
        "name": "log_{label}_strings",
        "description": "",
        "parameters": {
            "type": "object",
            "properties": {
                "identified_{label}_strings": {
                    "type": "array",
                    "description": "{description}",
                    "items": {
                        "type": "string",
                    },
                },
            },
        },
        "required": ["log_{label}_strings"],
    },
]


def extract_toks_from_batch(
    tok_batch: list[str], label: str, description: str, model="gpt-4o-mini"
):
    out = query_openai(
        messages=[
            {
                "role": "system",
                "content": prompt_template.format(
                    label=label, description=description, batch=tok_batch
                ),
            }
        ],
        model=model,
        functions=format_nest(functions_template, label=label, description=description),
        function_call={"name": f"log_{label}_strings"},
        temperature=0.7,
    )

    try:
        fn_call_args_str = out.function_call.arguments
    except Exception as e:
        print(f"Exception: {e}")
        print("No function call found.")
    try:
        fn_call_args = json.loads(fn_call_args_str)
    except Exception as e:
        print(f"Exception: {e}")
        print(f"Failed to load with json: {fn_call_args_str}")
        return []
    if f"identified_{label}_strings" not in fn_call_args:
        print(f"`identified_{label}_strings` not found in fn_call_args: {fn_call_args}")
        return []
    candidate_toks = fn_call_args[f"identified_{label}_strings"]

    return [tok for tok in candidate_toks if tok in tok_batch]


def create_label(
    label,
    desc,
    strs,
    batch_size=50,
    model="gpt-4o-mini",
    filters: Optional[list[str]] = None,
    stripped_labelling=True,
    n_passes: int = 1,
    inclusion_threshold: int = 1,
):
    """
    label: str = snake_case short representation of the label
    desc: str = description of strings that fit the label.
    strs: list of strs
    filters: list[str] = list of flags used to filter strings before labelling
        Available filters are ['space', 'no_space', 'capitalized', 'no_capitalized']
    stripped_labelling: bool = whether to strip strings before feeding to the language model.
    n_passes: int = number of times the LM sees the string.
    inclusion_threshold: int = the number of times the LM thinks the string fits the description before it's included in the final set (see n_passes argument to this fn)

    EG
    create_tok_label(label='past_tense_verb', desc='Strings that are past-tense verbs.')
    create_tok_label(label='adjective', desc='Strings that are adjectives.')
    """
    assert_snake_case(label)
    strs = list(strs)

    assert 1 <= inclusion_threshold <= n_passes

    filters = [] if filters is None else filters

    if len(filters) > 0:
        assert isinstance(filters, list)
        assert all([isinstance(k, str) for k in filters])
        assert all(
            [
                flag in {"space", "no_space", "capitalized", "no_capitalized"}
                for flag in filters
            ]
        )
        assert not (
            "space" in filters and "no_space" in filters
        ), "mutually exclusive filters options"
        assert not (
            "capitalized" in filters and "no_capitalized" in filters
        ), "mutually exclusive filters options"
        assert not (
            stripped_labelling is True and "no_space" in filters
        ), "stripped_labelling wouldn't change fn behavior here, so if you'd like no_space please set stripped_labelling to False. This assertion here is to reduce complexity."

    ws_path = tokre.get_workspace()
    # ls = os.listdir(ws_path)
    # if label+'.json' in ls:
    #     with open(f'{ws_path}/{label}.json', 'r') as f:
    #         existing_desc = json.load(f).get('create_tok_label_kwargs', {}).get('desc', 'No description found.')
    #     inp = input(f'{label}.json already found in {ws_path}/ with tok label description:\n{existing_desc}\nOverwrite this file?')
    #     if inp.strip().lower() not in {'y', 'yes'}:
    #         print('Aborting tok labelling.')
    #         return
    #     else:
    #         pass

    all_strs = strs
    from copy import deepcopy

    valid_strs = deepcopy(strs)

    if "space" in filters:
        valid_strs = extract_space_toks(valid_strs)
    if "no_space" in filters:
        valid_strs = extract_no_space_toks(valid_strs)
    if "capitalized" in filters:
        valid_strs = extract_capitalized_toks(valid_strs)
    if "no_capitalized" in filters:
        valid_strs = extract_not_capitalized_toks(valid_strs)

    if stripped_labelling is True:
        valid_strs = extract_stripped_toks(valid_strs)

    counts = np.zeros(len(all_strs), dtype=int)

    for pass_idx in range(n_passes):
        random.shuffle(valid_strs)
        batches = [
            valid_strs[i : i + batch_size]
            for i in range(0, len(valid_strs), batch_size)
        ]
        pbar_desc = f"{label}, pass {pass_idx+1}/{n_passes}"
        identified_strings = list(
            itertools.chain.from_iterable(
                threaded_map(
                    extract_toks_from_batch,
                    batches,
                    pbar_desc=pbar_desc,
                    n_threads=10,
                    kwargs={"label": label, "description": desc, "model": model},
                )
            )
        )
        for string in identified_strings:
            if stripped_labelling is True:
                if "no_space" not in filters and " " + string in all_strs:
                    counts[all_strs.index(" " + string)] += 1

                if "space" not in filters and string in all_strs:
                    counts[all_strs.index(string)] += 1
            else:
                if string in all_strs:
                    counts[all_strs.index(string)] += 1

    identified_str_indices = np.nonzero((counts >= inclusion_threshold).astype(int))[0]
    identified_strs = np.array(all_strs)[identified_str_indices].tolist()
    escaped_identified_strs = [tokre.escape(s) for s in identified_strs]

    literal_set = [tokre.tok_split(s) for s in identified_strs]

    fname = ws_path / f"{label}.json"

    result = {
        "create_label_kwargs": {
            "label": label,
            "desc": desc,
            "filters": filters,
            "stripped_labelling": stripped_labelling,
            "n_passes": n_passes,
            "inclusion_threshold": inclusion_threshold,
        },
        "literal_toks": literal_set,
        "literal_strs": identified_strs,
        "pattern": r"(" + "|".join(escaped_identified_strs) + r")",
    }

    save_dict(result, fname)

    return identified_strs



from tokre.core.modules import (
    Toks,
    Repeat,
    Phrase,
    OrGroup,
    VarDefn,
    VarRef,
    Wildcard,
    LearnedConst,
    Lookbehind,
    Lookahead,
)
from tokre.core.macros import DEFINED_MACROS
from tokre.core.parsing import parse
from lark import Transformer
import tokre


class InsertModules(Transformer):
    def repeat(self, children):
        assert len(children) == 3
        child_matcher, repeat_min, repeat_max = children
        return Repeat(child_matcher=child_matcher, min=repeat_min, max=repeat_max)

    def string(self, children):
        assert len(children) == 1, children
        assert isinstance(children[0], str)
        
        toks = [tokre.dec([tok_id]) for tok_id in list(tokre.enc(children[0]))]
        return Toks(toks=toks)

    def wildcard(self, children):
        return Wildcard()

    def phrase(self, children):
        return Phrase(matchers=children)

    def or_pattern(self, children):
        return OrGroup(matchers=children)

    def var_defn(self, children):
        assert len(children) == 2
        var_name, child_matcher = children
        return VarDefn(var_name=var_name, child_matcher=child_matcher)

    def var_ref(self, children):
        assert len(children) == 1, children
        var_name = children[0]
        return VarRef(var_name=var_name)

    def lookaround(self, children):
        child_module, is_backward, is_neg = children
        if is_backward is True:
            return Lookbehind(child_module, is_neg=is_neg)
        else:
            assert is_backward is False
            return Lookahead(child_module, is_neg=is_neg)

    def macro(self, children):
        assert len(children) >= 1

        macro_name, children = children[0], children[1:]

        args, kwargs = [], {}

        for child in children:
            if isinstance(child, dict):
                kwargs = kwargs | child
            else:
                args.append(child)

        if macro_name in DEFINED_MACROS:
            return DEFINED_MACROS[macro_name](*args, **kwargs)
        else:
            assert False, f"macro {macro_name} not found in macros.py"


def tree_to_module(tree):
    module = InsertModules().transform(tree)
    return module


from torch import nn


def recursively_add_name_to_submodule(module):
    assert not hasattr(
        module, "name_to_submodule"
    ), "module already has name_to_submodule attribute"
    name_to_submodule = {}

    # [STUB] hard to read code
    def add_named_submodules(module, name_to_submodule):
        for submodule in module.children():
            if hasattr(submodule, "name"):
                assert (
                    submodule.name not in name_to_submodule
                ), "Two tokre module children seem to have the same name?"
                name_to_submodule[submodule.name] = submodule

            if isinstance(submodule, nn.ModuleList):
                add_named_submodules(submodule, name_to_submodule)

        return name_to_submodule

    add_named_submodules(module, name_to_submodule)

    module.name_to_submodule = name_to_submodule
    for submodule in module.children():
        recursively_add_name_to_submodule(submodule)


def compile(s):
    tree = parse(s)
    module = tree_to_module(tree)
    if len(list(module.parameters())) == 0:
        module = LearnedConst(module)
    recursively_add_name_to_submodule(module)
    return module



import os

import backoff
from openai import OpenAI
import logging
import traceback


import tokre

logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger(__name__)


@backoff.on_exception(backoff.expo, Exception, max_time=120)
def query_openai(messages, model: str = "gpt-4o", temperature: float = 0.8, **kwargs):
    if isinstance(messages, str):
        messages = [{"role": "user", "content": messages}]

    if tokre._openai_api_key is not None:
        api_key = tokre._openai_api_key
    elif "OPENAI_API_KEY" in os.environ:
        api_key = os.environ["OPENAI_API_KEY"]
    else:
        raise ValueError(
            "Either the `OPENAI_API_KEY` environment variable needs to be set or `api_key` needs to be passed to `tokre.setup`."
        )
    client = OpenAI(api_key=api_key)
    try:
        response = client.chat.completions.create(
            model=model, messages=messages, temperature=temperature, **kwargs
        )
    except Exception as e:
        logger.error(f"Error in OpenAI API call: {str(e)}")
        logger.error(f"Full exception: {traceback.format_exc()}")
        raise
    return response.choices[0].message



from typing import Union
import tokre

def save_literal_set(label, desc, literals: list[Union[tuple[str], str]], vary_leading_space=False, vary_capitalization=False, strip_and_lowercase=False, metadata=None):
    if metadata is None:
        metadata = {}
    assert not ((vary_leading_space or vary_capitalization) and strip_and_lowercase)

    if strip_and_lowercase is True:
        literals = literals if isinstance(literals[0], str) else [''.join(toks) for toks in literals]
        literals = [literal.strip().lowercase() for literal in literals]

    if isinstance(literals[0], (tuple, list)):
        literal_strs = [''.join(literal) for literal in literals]
        literal_toks = literals
    elif isinstance(literals[0], str):
        literal_strs = literals
        literal_toks = [tokre.tok_split(literal_str) for literal_str in literal_strs]

    normalized_literals = {literal_str.strip().lower() for literal_str in literal_strs}

    if vary_leading_space is True or vary_capitalization is True:
        if vary_capitalization:
            normalized_literals = set([literal.capitalize() for literal in normalized_literals] + [literal for literal in normalized_literals])
        
        literal_toks = {tuple(tokre.tok_split(literal)) for literal in normalized_literals}
        
        if vary_leading_space:
            literal_toks = literal_toks.union({tuple(tokre.tok_split(' '+literal)) for literal in normalized_literals})
        
        literal_toks = [literal for literal in literal_toks if '[UNK]' not in literal]
        literal_strs = [''.join(literal) for literal in literal_toks]


    result = {
            "save_literal_kwargs": {
                "label": label,
                "desc": desc,
                "vary_leading_set": vary_leading_space,
                "vary_capitalization": vary_capitalization,
                "strip_and_lowercase": strip_and_lowercase,
            },
            "literal_toks": literal_toks,
            "literal_strs": literal_strs,
            "pattern": r"(" + "|".join(literal_strs) + r")",
            "metadata": metadata
        }

    ws = tokre.get_workspace()
    tokre.utils.save_dict(result, ws / (label + '.json'))

import json

def load_literal_set(label):
    ws = tokre.get_workspace()
    file_path = ws / (label + '.json')
    
    if not file_path.exists():
        raise FileNotFoundError(f"No literal set found for label '{label}'")
    
    with open(file_path, 'r') as file:
        result = json.load(file)
    
    return result


import torch.nn as nn
import tokre
from tokre.core.modules import PartialMatch, EmbedData
from tokre.core.parsing import parse
from tokre.core.tree_to_module import compile

from schedulefree import AdamWScheduleFree
import numpy as np
import torch
from frozendict import frozendict

from typing import Iterable

# import ray
from tqdm import tqdm

# Initialize Ray, using most available CPUs and ignoring reinit errors
# import multiprocessing
# num_cpus = max(1, multiprocessing.cpu_count() - 1)  # Leave one CPU free
# ray.init(ignore_reinit_error=True)#, _temp_dir="/media/noa/nvme_ssd/tmp")

def collect_matches(module, toks, aggr="longest"):
    assert aggr in ["shortest", "longest"]
    starting_matches = [
        PartialMatch(
            name="start",
            start=start_idx,
            end=start_idx,
            defns=frozendict(),
            data=None,
        )
        for start_idx in range(len(toks))
    ]

    unaggregated_matches = [
        match
        for start_match in starting_matches
        for match in module.matches(toks, start_match, reversed=False)
    ]

    end_to_aggr_match = {}
    for match in unaggregated_matches:
        if match.end in end_to_aggr_match:
            aggr_match = end_to_aggr_match[match.end]
            if len(match) > len(aggr_match):
                end_to_aggr_match[match.end] = match
        else:
            end_to_aggr_match[match.end] = match

    return list(end_to_aggr_match.values())


def is_int_or_tuple_of_ints(data):
    return isinstance(data, int) or (
        isinstance(data, tuple) and all([isinstance(x, int) for x in data])
    )


from tokre.core.modules import is_pred_data, Embed


def pred(module, match_data):
    assert is_pred_data(match_data) or (
        isinstance(module, Embed)
        and isinstance(match_data, int)
        or isinstance(match_data, tuple)
    )
    assert hasattr(module, "name")

    if isinstance(module, Embed):
        assert (
            isinstance(match_data, int)
            or isinstance(match_data, tuple)
            and all([isinstance])
        )
        return module.embed[match_data]

    if isinstance(match_data, list):
        assert hasattr(module, "mixer"), module
        preds = [torch.tensor(1.0)] + [pred(module, data) for data in match_data]
        preds = torch.stack(preds)

        return module.mixer(preds)

    elif isinstance(match_data, PartialMatch) or isinstance(match_data, EmbedData):
        match = match_data

        assert hasattr(module, "name_to_submodule")
        assert match.name in module.name_to_submodule, (module.name, match.name)
        return pred(module.name_to_submodule[match.name], match.data)

    elif match_data is None:
        return torch.tensor(1.0)

    else:
        raise ValueError("Unexpected match_data", match_data)


# @ray.remote(num_cpus=1)
# class ParallelModule:
#     def __init__(self, module, aggr, tokenizer):
#         self.module = module
#         self.aggr = aggr
#         print('initializing parallel module')
#         import tokre
#         tokre.setup(tokenizer=tokenizer)
#         print('initialized')

#     def get_matches(self, docs):
#         assert not isinstance(docs[0], str) and isinstance(docs[0], Iterable)
#         matches = [
#             collect_matches(self.module, toks=doc, aggr=self.aggr) for doc in docs
#         ]
#         return matches


class SynthFeat(nn.Module):
    def __init__(self, script, aggr="longest", n_actors=10, batch_size=100):
        super().__init__()
        assert aggr in ["longest", "shortest"]
        self.module = tokre.compile(script)
        self.aggr = aggr
        self.optimizer = AdamWScheduleFree(self.module.parameters(), lr=1e-3)

        self.batch_size = batch_size

    def get_matches(self, toks: list[str], n_matchers=1):
        assert n_matchers = 1          
        if isinstance(toks[0], list) or (
            isinstance(toks, np.ndarray) and len(toks.shape) == 2
        ):
            # toks is a list of documents
            docs = toks
            # if n_matchers > 1:
            #     print('creating parallel matchers')
            #     # module_ref = ray.put(self.module)
            #     # tokenizer_ref = ray.put(tokre.get_tokenizer())
            #     # tokre_ref = ray.put(tokre)
            #     parallel_matchers = [
            #         ParallelModule.remote(module_ref, self.aggr, tokenizer_ref) for _ in range(n_matchers)
            #     ]
            #     print('created parallel matchers')
            #     batched_docs = [
            #         docs[i : i + self.batch_size]
            #         for i in range(0, len(docs), self.batch_size)
            #     ]
            #     print('created batches')
            #     batched_matches = []
            #     K = 1  # Number of batches to process before waiting
            #     for i in range(0, len(batched_docs), K * len(parallel_matchers)):
            #         batch_futures = [
            #             parallel_matchers[j % len(parallel_matchers)].get_matches.remote(batched_docs[i + j])
            #             for j in range(min(K * len(parallel_matchers), len(batched_docs) - i))
            #         ]
            #         batched_matches.extend(ray.get(batch_futures))
            #     print('got matches')
            #     per_doc_matches = [match for batch in batched_matches for match in batch]
            #     return per_doc_matches
            # else:
            pbar = tqdm(docs, desc='collecting matches')
            per_doc_matches = [collect_matches(self.module, toks=doc, aggr=self.aggr) for doc in pbar]
            return per_doc_matches

        matches = collect_matches(self.module, toks=toks, aggr=self.aggr)
        return matches

    def get_mask(self, toks: Iterable, n_matchers=1):
        if isinstance(toks[0], Iterable) and not isinstance(toks[0], str):
            # toks is a list of documents
            assert all([isinstance(tok, str) for tok in toks[0]])
            matches = self.get_matches(toks, n_matchers=n_matchers)
            mask = torch.zeros((len(toks), len(toks[0])))
            for doc_idx, doc_matches in enumerate(matches):
                for match in doc_matches:
                    mask[doc_idx, match.end - 1] = 1.0
            return mask
        else:
            matches = self.get_matches(toks, n_matchers=1)
            mask = torch.zeros(len(toks))
            for match in matches:
                mask[match.end - 1] = 1.0

        return mask

    @torch.no_grad()
    def get_acts(self, toks, n_matchers=1):
        if isinstance(toks, Iterable) and isinstance(toks[0], str):
            synth_acts = torch.zeros(len(toks))
            matches = self.get_matches(toks, n_matchers=n_matchers)
            for match in matches:
                prediction = pred(self.module, match.data)
                synth_acts[match.end - 1] = prediction
        else:
            assert isinstance(toks, Iterable)
            assert isinstance(toks[0], Iterable)
            # return torch.stack([self.get_acts(doc) for doc in toks], dim=0)
            synth_acts = torch.zeros((len(toks), len(toks[0])))
            doc_matches = self.get_matches(toks, n_matchers=n_matchers)
            for doc_idx, matches in enumerate(doc_matches):
                for match in matches:
                    with torch.no_grad():
                        prediction = pred(self.module, match.data)
                        synth_acts[doc_idx, match.end-1] = prediction
        
        return synth_acts


    def train(self, toks, acts, parallel=True):
        from noa_tools import see
        print("getting matches")
        all_matches = self.get_matches(toks, parallel=parallel)
        print("training")
        for doc_matches, doc_acts in tqdm(zip(all_matches, acts)):
            for match in doc_matches:
                act = doc_acts[match.end - 1]
            
                loss = (pred(self.module, match.data) - act)**2

                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

    @property
    def pyregex(self):
        return self.module.pyregex



from tokre import SynthFeat
from typing import Iterable
import torch
import tokre
from tqdm import tqdm


def get_word_counts(dataset):
    '''
    Takes a transformers dataset with a text column, splits up each string into token strings using the tokre-setup tokenizer, and extracts word counts.
    '''
    word_synth = tokre.SynthFeat(
        r"""
valid_last_tok = (?<![re `[.,?!"';:\#$^&*()-]` search=True])
space_tok = [re ` [\S]+`][valid_last_tok]
nospace_tok = [re `[\S]+`][valid_last_tok]
capitalized_nospace_tok = [re `[A-Z].*`]

nospace_word = (?<=(\n)|[re `["]` search=True])[capitalized_nospace_tok][nospace_tok]*(?=[space_tok])
space_word = [space_tok][nospace_tok]*(?=[space_tok])

word = [nospace_word] | [space_word]

[word]
"""
    )

    word_counts = {}

    for item in tqdm(dataset):
        doc = item['text']
        try:
            toks = tokre.tok_split(doc)
        except Exception as e:
            print(f"Error tokenizing document: {e}")
            continue
        matches = word_synth.get_matches(toks)
        words = [
            tuple(toks[match.start : match.end])
            for match in matches
        ]

        for word in words:
            if word in word_counts:
                word_counts[word] += 1
            else:
                word_counts[word] = 1

    words, counts = list(word_counts.keys()), torch.tensor(list(word_counts.values()))

    perm = counts.topk(k=len(counts)).indices

    words = [words[i] for i in perm]
    counts = counts[perm].tolist()

    word_counts = [(word, count) for word, count in zip(words, counts)]

    return word_counts